@@ -22,20 +22,29 @@
 # TODO:
 #  - better annotate algorithm on packs
 
-import sys, time
+import sys, time, hashlib
+from itertools import groupby
 from PyQt4 import QtCore, QtGui
 
+from bzrlib.revision import CURRENT_REVISION
+
 from bzrlib.plugins.qbzr.lib.i18n import gettext
 from bzrlib.plugins.qbzr.lib.util import (
     BTN_CLOSE,
     QBzrWindow,
     ToolBarThrobberWidget,
     get_apparent_author_name,
-    get_set_encoding,
-    runs_in_loading_queue,
     get_icon,
-    FindToolbar,
+    get_monospace_font,
+    get_set_encoding,
+    get_set_tab_width_chars,
+    get_tab_width_pixels,
+    runs_in_loading_queue,
     )
+from bzrlib.plugins.qbzr.lib.widgets.toolbars import FindToolbar
+from bzrlib.plugins.qbzr.lib.widgets.texteditaccessory import (
+    GuideBarPanel, setup_guidebar_for_find
+)
 from bzrlib.plugins.qbzr.lib.uifactory import ui_current_widget
 from bzrlib.plugins.qbzr.lib.trace import reports_exception
 from bzrlib.plugins.qbzr.lib.logwidget import LogList
@@ -45,14 +54,16 @@ from bzrlib.plugins.qbzr.lib.texteditannotate import (AnnotateBarBase,
                                                       AnnotateEditerFrameBase)
 from bzrlib.lazy_import import lazy_import
 lazy_import(globals(), '''
+from bzrlib.config import parse_username
 from bzrlib.workingtree import WorkingTree
 from bzrlib.revisiontree import RevisionTree
-from bzrlib.revision import CURRENT_REVISION
 from bzrlib.plugins.qbzr.lib.revisionmessagebrowser import LogListRevisionMessageBrowser
 from bzrlib.plugins.qbzr.lib.encoding_selector import EncodingMenuSelector
+from bzrlib.plugins.qbzr.lib.widgets.tab_width_selector import TabWidthMenuSelector
 from bzrlib.plugins.qbzr.lib.syntaxhighlighter import highlight_document
 from bzrlib.plugins.qbzr.lib.revtreeview import paint_revno, get_text_color
 from bzrlib.plugins.qbzr.lib import logmodel
+from bzrlib.plugins.qbzr.lib.loggraphviz import BranchInfo
 from bzrlib.patiencediff import PatienceSequenceMatcher as SequenceMatcher
 ''')
 
@@ -64,7 +75,8 @@ class AnnotateBar(AnnotateBarBase):
         self.get_revno = get_revno
         self.annotate = None
         self.rev_colors = {}
-        self.highlight_revids = set()
+        self._highlight_revids = set()
+        self.highlight_lines = []
         
         self.splitter = None
         self.adjustWidth(1, 999)
@@ -74,6 +86,31 @@ class AnnotateBar(AnnotateBarBase):
             self.edit_cursorPositionChanged)
         self.show_current_line = False
 
+    def get_highlight_revids(self):
+        return self._highlight_revids
+
+    def set_highlight_revids(self, value):
+        if self._highlight_revids == value:
+            return
+
+        self._highlight_revids = value
+        self.update_highlight_lines()
+
+    def update_highlight_lines(self):
+        self.highlight_lines = []
+        if not self.annotate:
+            return
+        lines = [i for i, (revno, istop) in enumerate(self.annotate)
+                 if revno in self._highlight_revids]
+
+        # Convert [0,1,2,5,6,9,14,15,16,17] to [(0,3),(5,2),(9,1),(14,4)]
+        def summarize(lines):
+            for k, g in groupby(enumerate(lines), key=lambda x:x[1]-x[0]):
+                yield [line for i, line in g]
+        self.highlight_lines = [(x[0], len(x)) for x in summarize(lines)]
+
+    highlight_revids = property(get_highlight_revids, set_highlight_revids)
+
     def edit_cursorPositionChanged(self):
         self.show_current_line = True
 
@@ -132,7 +169,7 @@ class AnnotateBar(AnnotateBarBase):
         if self.annotate and line_number-1 < len(self.annotate):
             revid, is_top = self.annotate[line_number - 1]
             if is_top:
-                if revid in self.highlight_revids:
+                if revid in self._highlight_revids:
                     font = painter.font()
                     font.setBold(True)
                     painter.setFont(font)
@@ -228,21 +265,26 @@ class AnnotatedTextEdit(QtGui.QPlainTextEdit):
 
 class AnnotateWindow(QBzrWindow):
 
-    def __init__(self, branch, tree, path, fileId, encoding=None, parent=None,
-                 ui_mode=True, loader=None, loader_args=None, no_graph=False):
+    def __init__(self, branch, working_tree, annotate_tree, path, fileId,
+                 encoding=None, parent=None, ui_mode=True, no_graph=False,
+                 loader=None, loader_args=None, activate_line=None):
         QBzrWindow.__init__(self,
                             [gettext("Annotate"), gettext("Loading...")],
                             parent, ui_mode=ui_mode)
         self.restoreSize("annotate", (780, 680))
 
+        self.activate_line_after_load = activate_line
+
         self.windows = []
 
         self.branch = branch
-        self.tree = tree
-        if isinstance(tree, WorkingTree):
-            self.working_tree = tree
-        else:
-            self.working_tree = None
+        self.annotate_tree = annotate_tree
+        self.working_tree = working_tree
+        if (self.working_tree is None and 
+            isinstance(annotate_tree, WorkingTree)):
+            self.working_tree = annotate_tree
+        
+        self.no_graph = no_graph
         self.old_lines = None
         
         self.fileId = fileId
@@ -261,14 +303,14 @@ class AnnotateWindow(QBzrWindow):
             QtCore.Qt.TextSelectableByKeyboard)
         self.text_edit.setLineWrapMode(QtGui.QPlainTextEdit.NoWrap)
         
-        self.text_edit.document().setDefaultFont(
-            QtGui.QFont("Courier New,courier", 
-                        self.text_edit.font().pointSize()))
-        
+        self.text_edit.document().setDefaultFont(get_monospace_font())
+
+        self.guidebar_panel = GuideBarPanel(self.text_edit, parent=self)
+        self.guidebar_panel.add_entry('annotate', QtGui.QColor(255, 160, 180))
         self.annotate_bar = AnnotateBar(self.text_edit, self, self.get_revno)
         annotate_spliter = QtGui.QSplitter(QtCore.Qt.Horizontal, self)
         annotate_spliter.addWidget(self.annotate_bar)
-        annotate_spliter.addWidget(self.text_edit)
+        annotate_spliter.addWidget(self.guidebar_panel)
         self.annotate_bar.splitter = annotate_spliter
         self.text_edit_frame.hbox.addWidget(annotate_spliter)
         
@@ -277,9 +319,12 @@ class AnnotateWindow(QBzrWindow):
                      self.edit_cursorPositionChanged)        
         self.connect(self.annotate_bar,
                      QtCore.SIGNAL("cursorPositionChanged()"),
-                     self.edit_cursorPositionChanged)        
+                     self.edit_cursorPositionChanged)
+        self.connect(self.text_edit,
+                     QtCore.SIGNAL("documentChangeFinished()"),
+                     self.edit_documentChangeFinished)
         
-        self.log_list = AnnotateLogList(self.processEvents, self.throbber, no_graph, self)
+        self.log_list = AnnotateLogList(self.processEvents, self.throbber, self)
         self.log_list.header().hideSection(logmodel.COL_DATE)
         self.log_list.parent_annotate_window = self
         self.log_branch_loaded = False
@@ -330,7 +375,15 @@ class AnnotateWindow(QBzrWindow):
         self.connect(word_wrap,
                      QtCore.SIGNAL("toggled (bool)"),
                      self.word_wrap_toggle)
+
+        def setTabStopWidth(tw):
+            self.text_edit.setTabStopWidth(get_tab_width_pixels(tab_width_chars=tw))
+            get_set_tab_width_chars(branch=self.branch,tab_width_chars=tw)
+        self.tab_width_selector = TabWidthMenuSelector(get_set_tab_width_chars(branch=branch),
+                gettext("Tab Width"),
+                setTabStopWidth)
         
+        view_menu.addMenu(self.tab_width_selector)
         view_menu.addMenu(self.encoding_selector)
         view_menu.addAction(word_wrap)
         
@@ -357,6 +410,7 @@ class AnnotateWindow(QBzrWindow):
         self.connect(self.show_find,
                      QtCore.SIGNAL("toggled (bool)"),
                      self.show_find_toggle)
+        setup_guidebar_for_find(self.guidebar_panel, self.find_toolbar, index=1)
         
         self.goto_line_toolbar = GotoLineToolbar(self, self.show_goto_line)
         self.goto_line_toolbar.hide()
@@ -366,6 +420,8 @@ class AnnotateWindow(QBzrWindow):
                      QtCore.SIGNAL("toggled (bool)"),
                      self.show_goto_line_toggle )
 
+        self.__hashes = {}
+
     def show(self):
         QBzrWindow.show(self)
         QtCore.QTimer.singleShot(0, self.initial_load)
@@ -382,7 +438,7 @@ class AnnotateWindow(QBzrWindow):
         try:
             if self.loader_func is not None:
                 (self.branch,
-                 self.tree,
+                 self.annotate_tree,
                  self.working_tree,
                  self.path, self.fileId) = self.loader_func(*self.loader_args)
                 self.loader_func = self.loader_args = None # kill extra refs...
@@ -392,16 +448,33 @@ class AnnotateWindow(QBzrWindow):
             self.branch.lock_read()
             try:
                 self.set_annotate_title()
-                self.annotate(self.tree, self.fileId, self.path)
+                self.annotate(self.annotate_tree, self.fileId, self.path)
             finally:
                 self.branch.unlock()
         finally:
             self.throbber.hide()
-    
+        QtCore.QTimer.singleShot(1, self._activate_line)
+
+    def _activate_line(self):
+        # [bialix 2012/09/14] I have to use QtCore.QTimer.singleShot because
+        # otherwise the line with uncommitted changes is not highlighted
+        # properly: nothing appears in the revision message browser at all.
+        # I didn't find the reason behind it, shame on me.
+        n = self.activate_line_after_load
+        self.activate_line_after_load = None    # clear the line number just in case
+        if n:
+            if n == 1:
+                # [bialix 2012/09/13] actually after loading annotation
+                # document implictly sets current position to the first line.
+                # So we're unable to change position to 1st line.
+                # Make a hack: select 2nd line, and then actually select 1st one.
+                self.go_to_line(2)
+            self.go_to_line(n)
+
     def set_annotate_title(self):
         # and update the title to show we are done.
-        if isinstance(self.tree, RevisionTree):
-            revno = self.get_revno(self.tree.get_revision_id())
+        if isinstance(self.annotate_tree, RevisionTree):
+            revno = self.get_revno(self.annotate_tree.get_revision_id())
             if revno:
                 self.set_title_and_icon([gettext("Annotate"), self.path,
                                          gettext("Revision %s") % revno])
@@ -410,21 +483,25 @@ class AnnotateWindow(QBzrWindow):
         self.set_title_and_icon([gettext("Annotate"), self.path])
 
     def get_revno(self, revid):
-        if revid in self.log_list.graph_provider.revid_rev:
-            return self.log_list.graph_provider.revid_rev[revid].revno_str
+        gv = self.log_list.log_model.graph_viz
+        if (gv and
+            revid in gv.revid_rev):
+            return gv.revid_rev[revid].revno_str
         return ""
     
-    def annotate(self, tree, fileId, path):
+    def annotate(self, annotate_tree, fileId, path):
         self.now = time.time()
         self.rev_indexes = {}
         last_revid = None
         lines = []
         annotate = []
         ordered_revids = []
-        
-        
+
         self.processEvents()
-        for revid, text in tree.annotate_iter(fileId):
+        for revid, text in annotate_tree.annotate_iter(fileId):
+            if revid == CURRENT_REVISION:
+                revid = CURRENT_REVISION + annotate_tree.basedir
+            
             text = text.decode(self.encoding, 'replace')
             
             lines.append(text)
@@ -460,34 +537,33 @@ class AnnotateWindow(QBzrWindow):
         self.annotate_bar.annotate = annotate
         self.text_edit.annotate = annotate
         self.annotate_bar.show_current_line = False
+
+        self.text_edit.emit(QtCore.SIGNAL("documentChangeFinished()"))
         
         self.processEvents()
         
         just_loaded_log = False
         if not self.log_branch_loaded:
             self.log_branch_loaded = True
-            self.log_list.load_branch(self.branch, [self.fileId], tree)
+            bi = BranchInfo('', self.working_tree, self.branch)
+            self.log_list.load(
+                (bi,), bi, [self.fileId], self.no_graph,
+                logmodel.WithWorkingTreeGraphVizLoader)
+            
+            gv = self.log_list.log_model.graph_viz
+            self.annotate_bar.adjustWidth(len(lines),
+                                          gv.revisions[0].revno_sequence[0])
             
             just_loaded_log = True
             
-            # Show the revisions the we know about now.
-            gp = self.log_list.graph_provider
-            gp.filter_file_id = [False for i in xrange(len(gp.revisions))]
-            
-            changed_indexes = []
+            # Show the revisions the we know about from the annotate.
+            filter = self.log_list.log_model.file_id_filter
+            changed_revs = []
             for revid in self.rev_indexes.keys():
-                index = gp.revid_rev[revid].index
-                gp.filter_file_id[index] = True
-                changed_indexes.append(index)
-                
-                if len(changed_indexes) >=500:
-                    gp.invaladate_filter_cache_revs(changed_indexes)
-                    changed_indexes = []
-            
-            gp.invaladate_filter_cache_revs(changed_indexes, last_call=True)
-            
-            self.annotate_bar.adjustWidth(len(lines),
-                                          gp.revisions[0].revno_sequence[0])
+                rev = gv.revid_rev[revid]
+                filter.filter_file_id[rev.index] = True
+                changed_revs.append(rev)
+            filter.filter_changed_callback(changed_revs, last_call=True)
         
         self.processEvents()
         highlight_document(self.text_edit, path)
@@ -499,16 +575,11 @@ class AnnotateWindow(QBzrWindow):
         
         if just_loaded_log:
             # Check for any other revisions we don't know about
-            revids = [rev.revid for rev in gp.revisions
-                      if rev.revid not in self.rev_indexes]
             
-            for repo, revids in gp.get_repo_revids(revids):
-                chunk_size = 500
-                
-                for start in xrange(0, len(revids), chunk_size):
-                    gp.load_filter_file_id_chunk(repo, 
-                            revids[start:start + chunk_size])
-            gp.load_filter_file_id_chunk_finished()
+            filter = self.log_list.log_model.file_id_filter
+            revids = [rev.revid for rev in gv.revisions
+                      if rev.revid not in self.rev_indexes]
+            filter.load(revids)
     
     def translate_positions(self, old_lines, new_lines, old_positions):
         sm = SequenceMatcher(None, old_lines, new_lines)
@@ -541,39 +612,66 @@ class AnnotateWindow(QBzrWindow):
                 old_char_start += old_len
                 new_char_start += new_len
         return new_positions
-    
+
     def revisions_loaded(self, revisions, last_call):
         for rev in revisions.itervalues():
-            author_name = get_apparent_author_name(rev)
-            
+            authors = rev.get_apparent_authors()
+            emails = map(self._maybe_extract_email, authors)
+            author_id = ';'.join(emails)
+
             if rev.timestamp is None:
                 days = sys.maxint
             elif self.now < rev.timestamp:
                 days = 0
             else:
                 days = (self.now - rev.timestamp) / (24 * 60 * 60)
-            
+
             alpha = 0.5/((days/50) + 1)
-            hue =  1-float(abs(hash(author_name))) / sys.maxint 
+            h_sh = self._get_hash(author_id)
+            hue =  1-float(h_sh) / sys.maxint
             color = QtGui.QColor.fromHsvF(hue, 1, 1, alpha)
             brush = QtGui.QBrush(color)
-            
+
             self.annotate_bar.rev_colors[rev.revision_id] = brush
             self.text_edit.rev_colors[rev.revision_id] = brush
-        
+
         self.annotate_bar.update()
         self.text_edit.update()
 
+    def _maybe_extract_email(self, author):
+        name, email = parse_username(author)
+        if email:
+            return email
+        return author
+
+    def _get_hash(self, author_id):
+        h = self.__hashes.get(author_id)
+        if h is None:
+            h = abs(hash(author_id))
+            #h = abs(hash(hashlib.md5(author_id).hexdigest())) # maybe increase enthropy via md5?
+            # XXX [bialix 2012/03/30]
+            # I don't really like how we create colors for annotation.
+            # Also, abs(hash) loses 1 bit of the hash result.
+            # I think we can improve our algorithm to use more distinctive colors
+            # maybe we should use md5 as more enthropy source for hash,
+            # or maybe we can use some sort of predefined set of colors
+            # and each next author will use color from that set,
+            # but in this case we lose consistency of colors for the same author
+            # between different annotations
+            # I have no clever idea yet.
+            self.__hashes[author_id] = h
+        return h
+
     def edit_cursorPositionChanged(self):
         current_line = self.text_edit.document().findBlock(
             self.text_edit.textCursor().position()).blockNumber()
         if self.text_edit.annotate:
             rev_id, is_top = self.text_edit.annotate[current_line]
-            if self.log_list.graph_provider.has_rev_id(rev_id):
-                self.log_list.log_model.ensure_rev_visible(rev_id)
-                index = self.log_list.log_model.indexFromRevId(rev_id)
-                index = self.log_list.filter_proxy_model.mapFromSource(index)
-                self.log_list.setCurrentIndex(index)
+            self.log_list.select_revid(rev_id)
+
+    def edit_documentChangeFinished(self):
+        self.annotate_bar.update_highlight_lines()
+        self.guidebar_panel.update_data(annotate=self.annotate_bar.highlight_lines)
 
     @runs_in_loading_queue
     def set_annotate_revision(self):
@@ -582,14 +680,15 @@ class AnnotateWindow(QBzrWindow):
             self.branch.lock_read()
             try:
                 revid = str(self.log_list.currentIndex().data(logmodel.RevIdRole).toString())
-                if revid == CURRENT_REVISION:
-                    self.tree = self.working_tree
+                if revid.startswith(CURRENT_REVISION):
+                    rev = cached_revisions[revid]
+                    self.annotate_tree = self.working_tree
                 else:
-                    self.tree = self.branch.repository.revision_tree(revid)
-                self.path = self.tree.id2path(self.fileId)
+                    self.annotate_tree = self.branch.repository.revision_tree(revid)
+                self.path = self.annotate_tree.id2path(self.fileId)
                 self.set_annotate_title()
                 self.processEvents()
-                self.annotate(self.tree, self.fileId, self.path)
+                self.annotate(self.annotate_tree, self.fileId, self.path)
             finally:
                 self.branch.unlock()
         finally:
@@ -603,7 +702,7 @@ class AnnotateWindow(QBzrWindow):
         try:
             self.branch.lock_read()
             try:
-                self.annotate(self.tree, self.fileId, self.path)
+                self.annotate(self.annotate_tree, self.fileId, self.path)
             finally:
                 self.branch.unlock()
         finally:
@@ -612,6 +711,7 @@ class AnnotateWindow(QBzrWindow):
     def log_list_selectionChanged(self, selected, deselected):
         revids = self.log_list.get_selection_and_merged_revids()
         self.annotate_bar.highlight_revids = revids
+        self.guidebar_panel.update_data(annotate=self.annotate_bar.highlight_lines)
         self.annotate_bar.update()
     
     def show_find_toggle(self, state):
@@ -632,6 +732,14 @@ class AnnotateWindow(QBzrWindow):
         else:
             self.text_edit.setLineWrapMode(QtGui.QPlainTextEdit.NoWrap)
 
+    def go_to_line(self, line):
+        doc = self.text_edit.document()
+        cursor = QtGui.QTextCursor(doc)
+        cursor.setPosition(doc.findBlockByNumber(line-1).position())
+        self.text_edit.setTextCursor(cursor)
+        self.text_edit.centerCursor()
+
+
 # QIntValidator did not work on vila's setup, so this is a workaround.
 class IntValidator(QtGui.QValidator):
     def validate (self, input, pos):
@@ -646,6 +754,7 @@ class IntValidator(QtGui.QValidator):
         else:
             return (QtGui.QValidator.Invalid, pos)
 
+
 class GotoLineToolbar(QtGui.QToolBar):
     
     def __init__(self, anotate_window, show_action):
@@ -693,33 +802,18 @@ class GotoLineToolbar(QtGui.QToolBar):
     
     def go_triggered(self, state=True):
         try:
-            line = int(str(self.line_edit.text()))-1
+            line = int(str(self.line_edit.text()))
         except ValueError:
             pass
         else:
-            doc = self.anotate_window.text_edit.document()
-            cursor = QtGui.QTextCursor(doc)
-            cursor.setPosition(doc.findBlockByNumber(line).position())
-            self.anotate_window.text_edit.setTextCursor(cursor)
-            
+            self.anotate_window.go_to_line(line)
             self.show_action.setChecked(False)
 
 
-
 class AnnotateLogList(LogList):
     
     parent_annotate_window = None
     
-    def load(self):
-        self.graph_provider.lock_read_branches()
-        try:
-            self.graph_provider.load_tags()
-            self.log_model.load_graph(
-                self.graph_provider.load_graph_all_revisions_for_annotate)
-            self._adjust_revno_column()
-        finally:
-            self.graph_provider.unlock_branches()
-    
     def create_context_menu(self):
         LogList.create_context_menu(self, diff_is_default_action=False)
         set_rev_action = QtGui.QAction(gettext("&Annotate this revision"),
