@@ -18,6 +18,8 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+import signal
+
 from bzrlib import errors
 from bzrlib.commands import Command
 from bzrlib.option import Option
@@ -33,6 +35,7 @@ from bzrlib import (
     builtins,
     osutils,
     ui,
+    gpg,
     )
 from bzrlib.branch import Branch
 from bzrlib.bzrdir import BzrDir
@@ -57,6 +60,8 @@ from bzrlib.plugins.qbzr.lib.log import LogWindow
 from bzrlib.plugins.qbzr.lib.info import QBzrInfoWindow
 from bzrlib.plugins.qbzr.lib.init import QBzrInitWindow
 from bzrlib.plugins.qbzr.lib.main import QBzrMainWindow
+from bzrlib.plugins.qbzr.lib.verify_signatures import \
+QBzrVerifySignaturesWindow
 from bzrlib.plugins.qbzr.lib.pull import (
     QBzrPullWindow,
     QBzrPushWindow,
@@ -74,6 +79,8 @@ from bzrlib.plugins.qbzr.lib.util import (
     )
 from bzrlib.plugins.qbzr.lib.uifactory import QUIFactory
 from bzrlib.plugins.qbzr.lib.send import SendWindow
+from bzrlib.plugins.qbzr.lib.shelvewindow import ShelveWindow
+from bzrlib.plugins.qbzr.lib.widgets.shelvelist import ShelveListWindow
 ''')
 
 from bzrlib.plugins.qbzr.lib.diff_arg import DiffArgProvider
@@ -144,7 +151,7 @@ class QBzrCommand(Command):
        self.main_window should be instance of QBzrWindow or QBzrDialog
        with attribute "return_code" set to 0 or 1.
     """
-    
+
     @install_gettext
     @report_missing_pyqt
     def run(self, *args, **kwargs):
@@ -157,10 +164,13 @@ class QBzrCommand(Command):
         try:
             ui.ui_factory = QUIFactory()
             
+            # Handle interupt signal correctly.
+            signal.signal(signal.SIGINT, signal.SIG_DFL)
+            
             # Set up global exception handling.
             from bzrlib.plugins.qbzr.lib.trace import excepthook
             sys.excepthook = excepthook
-            
+
             try:
                 try:
                     ret_code = self._qbzr_run(*args, **kwargs)
@@ -172,7 +182,10 @@ class QBzrCommand(Command):
                 if ret_code is None:
                     main_window = getattr(self, "main_window", None)
                     if main_window is not None:
-                        ret_code = getattr(main_window, "return_code", None)
+                        # 0.20 special: We check hasattr() first to work around
+                        # <http://bugs.python.org/issue4230>
+                        if hasattr(main_window, "return_code"):
+                            ret_code = main_window.return_code
                 return ret_code
             except Exception:
                 ui_mode = kwargs.get("ui_mode", False)
@@ -189,7 +202,7 @@ execute_option = Option("execute", short_name='e',
          "waiting for user input.")
 
 # A special option so 'revision' can be passed as a simple string, when we do
-# *not* wan't bzrlib's feature of parsing the revision string before passing it.
+# *not* want bzrlib's feature of parsing the revision string before passing it.
 # This is used when we just want a plain string to pass to our dialog for it to
 # display in the UI, and we will later pass it to bzr for parsing. If you want
 # bzrlib to parse and pass a revisionspec object, just pass the string
@@ -221,9 +234,12 @@ class cmd_qannotate(QBzrCommand):
     takes_args = ['filename']
     takes_options = ['revision',
                      Option('encoding', type=check_encoding,
-                     help='Encoding of files content (default: utf-8).'),
+                         help='Encoding of files content (default: utf-8).'),
                      ui_mode_option,
                      Option('no-graph', help="Shows the log with no graph."),
+                     Option('line', short_name='L', type=int, argname='N',
+                        param_name='activate_line',
+                        help='Activate line N on start.'),
                     ]
     aliases = ['qann', 'qblame']
 
@@ -251,11 +267,12 @@ class cmd_qannotate(QBzrCommand):
             else:
                 tree = branch.repository.revision_tree(
                         revision_id = revision[0].in_history(branch).rev_id)
-            
+
             file_id = tree.path2id(relpath)
             if file_id is None:
                 raise errors.NotVersionedError(filename)
-            entry = tree.inventory[file_id]
+            [(path, entry)] = list(tree.iter_entries_by_dir(
+                specific_file_ids=[file_id]))
             if entry.kind != 'file':
                 raise errors.BzrCommandError(
                         'bzr qannotate only works for files (got %r)' % entry.kind)
@@ -274,12 +291,12 @@ class cmd_qannotate(QBzrCommand):
         return branch, tree, wt, relpath, file_id
 
     def _qbzr_run(self, filename=None, revision=None, encoding=None,
-                  ui_mode=False, no_graph=False):
-        win = AnnotateWindow(None, None, None, None,
+                  ui_mode=False, no_graph=False, activate_line=None):
+        win = AnnotateWindow(None, None, None, None, None,
                              encoding=encoding, ui_mode=ui_mode,
                              loader=self._load_branch,
                              loader_args=(filename, revision),
-                             no_graph=no_graph)
+                             no_graph=no_graph, activate_line=activate_line)
         win.show()
         self._application.exec_()
 
@@ -288,9 +305,9 @@ class cmd_qadd(QBzrCommand):
     """GUI for adding files or directories."""
     takes_args = ['selected*']
     takes_options = [ui_mode_option]
-    
+
     def _qbzr_run(self, selected_list=None, ui_mode=False):
-        tree, selected_list = builtins.tree_files(selected_list)
+        tree, selected_list = WorkingTree.open_containing_paths(selected_list)
         if selected_list == ['']:
             selected_list = None
         self.main_window = AddWindow(tree, selected_list, dialog=False, ui_mode=ui_mode)
@@ -304,7 +321,7 @@ class cmd_qrevert(QBzrCommand):
     takes_options = [ui_mode_option, bzr_option('revert', 'no-backup')]
 
     def _qbzr_run(self, selected_list=None, ui_mode=False, no_backup=False):
-        tree, selected_list = builtins.tree_files(selected_list)
+        tree, selected_list = WorkingTree.open_containing_paths(selected_list)
         if selected_list == ['']:
             selected_list = None
         self.main_window = RevertWindow(tree, selected_list, dialog=False,
@@ -334,6 +351,7 @@ class cmd_qbrowse(QBzrCommand):
     aliases = ['qbw']
 
     def _qbzr_run(self, revision=None, location=None):
+        Branch.open_containing(location or u'.')  # if there is no branch we want NotBranchError raised
         if revision is None:
             win = BrowseWindow(location = location)
         else:
@@ -348,12 +366,24 @@ class cmd_qcommit(QBzrCommand):
     takes_options = [
             bzr_option('commit', 'message'),
             bzr_option('commit', 'local'),
+            bzr_option('commit', 'file'),
+            Option('file-encoding', type=check_encoding,
+               help='Encoding of commit message file content.'),
             ui_mode_option,
             ]
     aliases = ['qci']
 
-    def _qbzr_run(self, selected_list=None, message=None, local=False, ui_mode=False):
-        tree, selected_list = builtins.tree_files(selected_list)
+    def _qbzr_run(self, selected_list=None, message=None, file=None,
+                  local=False, ui_mode=False, file_encoding=None):
+        if message is not None and file:
+            raise errors.BzrCommandError("please specify either --message or --file")
+        if file:
+            f = open(file)
+            try:
+                message = f.read().decode(file_encoding or osutils.get_user_encoding())
+            finally:
+                f.close()
+        tree, selected_list = WorkingTree.open_containing_paths(selected_list)
         if selected_list == ['']:
             selected_list = None
         self.main_window = CommitWindow(tree, selected_list, dialog=False,
@@ -375,6 +405,8 @@ class cmd_qdiff(QBzrCommand, DiffArgProvider):
         Option('modified', short_name='M',
                help='Show diff for modified files.'),
         Option('renamed', short_name='R', help='Show diff for renamed files.'),
+        Option('ignore-whitespace', short_name='w',
+               help="Ignore whitespace when finding differences"),
         bzr_option('diff', 'old'),
         bzr_option('diff', 'new'),
         ]
@@ -383,24 +415,26 @@ class cmd_qdiff(QBzrCommand, DiffArgProvider):
     aliases = ['qdi']
 
     def get_diff_window_args(self, processEvents, add_cleanup):
+        args = {}
         try:
             from bzrlib.diff import get_trees_and_branches_to_diff_locked
         except ImportError:
             from bzrlib.diff import get_trees_and_branches_to_diff
-            (old_tree, new_tree,
-             old_branch, new_branch,
-             specific_files, extra_trees) = \
+            (args["old_tree"], args["new_tree"],
+             args["old_branch"], args["new_branch"],
+             args["specific_files"], _) = \
                 get_trees_and_branches_to_diff(
                     self.file_list, self.revision, self.old, self.new)
         else:
-            (old_tree, new_tree,
-             old_branch, new_branch,
-             specific_files, extra_trees) = \
+            (args["old_tree"], args["new_tree"],
+             args["old_branch"], args["new_branch"],
+             args["specific_files"], _) = \
                 get_trees_and_branches_to_diff_locked(
                     self.file_list, self.revision, self.old, self.new,
                     add_cleanup)
-        return old_tree, new_tree, old_branch, new_branch, specific_files
-    
+        args["ignore_whitespace"] = self.ignore_whitespace
+        return args
+
     def get_ext_diff_args(self, processEvents):
         args = []
         if self.revision and len(self.revision) == 1:
@@ -413,14 +447,14 @@ class cmd_qdiff(QBzrCommand, DiffArgProvider):
             args.append("--new=%s" % self.new)
         if self.old and not self.old == CUR_DIR:
             args.append("--old=%s" % self.old)
-        
+
         if self.file_list:
             args.extend(self.file_list)
-        
-        return None, args    
+
+        return None, args
 
     def _qbzr_run(self, revision=None, file_list=None, complete=False,
-            encoding=None,
+            encoding=None, ignore_whitespace=False,
             added=None, deleted=None, modified=None, renamed=None,
             old=None, new=None, ui_mode=False):
 
@@ -433,11 +467,12 @@ class cmd_qdiff(QBzrCommand, DiffArgProvider):
         if not (added or deleted or modified or renamed):
             # if no filter option used then turn all on
             filter_options.all_enable()
-        
+
         self.revision = revision
         self.file_list = file_list
         self.old = old
         self.new = new
+        self.ignore_whitespace = ignore_whitespace
 
         window = DiffWindow(self,
                             complete=complete,
@@ -452,7 +487,7 @@ class cmd_qlog(QBzrCommand):
     """Show log of a repository, branch, file, or directory in a Qt window.
 
     By default show the log of the branch containing the working directory.
-    
+
     If multiple files are speciffied, they must be from the same branch.
 
     :Examples:
@@ -470,13 +505,17 @@ class cmd_qlog(QBzrCommand):
     """
 
     takes_args = ['locations*']
-    takes_options = [ui_mode_option,
-                   	 Option('no-graph', help="Shows the log with no graph."),
-                    ]
+    takes_options = [
+        ui_mode_option,
+        Option('no-graph', help="Shows the log with no graph."),
+        Option('show-trees', help="Show working trees that have changes "
+                                  "as nodes in the graph"),
+        ]
 
-    def _qbzr_run(self, locations_list=None, ui_mode=False, no_graph=False):
+    def _qbzr_run(self, locations_list=None, ui_mode=False, no_graph=False,
+                  show_trees=False):
         window = LogWindow(locations_list, None, None, ui_mode=ui_mode,
-                           no_graph=no_graph)
+                           no_graph=no_graph, show_trees=show_trees)
         window.show()
         self._application.exec_()
 
@@ -606,7 +645,7 @@ class cmd_qpush(QBzrCommand):
 
         if directory is None:
             directory = CUR_DIR
-        
+
         try:
             tree_to = WorkingTree.open_containing(directory)[0]
             branch_to = tree_to.branch
@@ -656,6 +695,29 @@ class cmd_qinfo(QBzrCommand):
         self._application.exec_()
 
 
+class cmd_qverify_signatures(QBzrCommand):
+    """Shows digital signature statuses for branch commits"""
+
+    takes_options = [
+            Option('acceptable-keys',
+                   help='Comma separated list of GPG key patterns which are'
+                        ' acceptable for verification.',
+                   short_name='k',
+                   type=str,),
+            'revision', 
+          ]
+    takes_args = ['location?']
+
+    def _qbzr_run(self, acceptable_keys=None, revision=None, location=CUR_DIR):
+        if gpg.GPGStrategy.verify_signatures_available():
+            window = QBzrVerifySignaturesWindow(acceptable_keys, revision,
+                                                                    location)
+            window.show()
+            self._application.exec_()
+        else:
+            raise errors.DependencyNotPresent("python-gpgme", 
+                                        "python-gpgme not installed")
+
 class cmd_qinit(QBzrCommand):
     """Initializes a new branch or shared repository."""
 
@@ -693,7 +755,7 @@ class cmd_merge(bzrlib.builtins.cmd_merge, DiffArgProvider):
         tree_merger = self.merger.make_merger()
         self.tt = tree_merger.make_preview_transform()
         result_tree = self.tt.get_preview_tree()
-        return self.merger.this_tree, result_tree, None, None, None
+        return {"old_tree": self.merger.this_tree, "new_tree": result_tree}
 
     @install_gettext
     @report_missing_pyqt
@@ -701,7 +763,7 @@ class cmd_merge(bzrlib.builtins.cmd_merge, DiffArgProvider):
         # Set up global execption handeling.
         from bzrlib.plugins.qbzr.lib.trace import excepthook
         sys.excepthook = excepthook
-        
+
         self.merger = merger
         try:
             window = DiffWindow(self, encoding=self._encoding)
@@ -749,12 +811,12 @@ class cmd_qmain(QBzrCommand):
 #     bzr qannotate commands.py -r1117
 
 class cmd_qsubprocess(Command):
-    """Run some bzr command as subprocess. 
+    """Run some bzr command as subprocess.
     Used with most of subprocess-based dialogs of QBzr.
-    
+
     If CMD argument starts with @ characters then it used as name of file with
     actual cmd string (in utf-8).
-    
+
     With --bencode option cmd string interpreted as bencoded list of utf-8
     strings. This is the recommended way to launch qsubprocess.
     """
@@ -797,7 +859,7 @@ class cmd_qgetnew(QBzrCommand):
     takes_options = [ui_mode_option]
     aliases = ['qgetn']
 
-    def _qbzr_run(self, location=CUR_DIR, ui_mode=False):
+    def _qbzr_run(self, location=None, ui_mode=False):
         from bzrlib.plugins.qbzr.lib.getnew import GetNewWorkingTreeWindow
         self.main_window = GetNewWorkingTreeWindow(location, ui_mode=ui_mode)
         self.main_window.show()
@@ -917,7 +979,7 @@ class cmd_qsend(QBzrCommand):
 
     takes_args = ['submit_branch?', 'public_branch?']
     takes_options = [ui_mode_option]
-    
+
     def _qbzr_run(self, submit_branch=CUR_DIR, public_branch=None, ui_mode=False):
         branch = Branch.open_containing(submit_branch)[0]
         window = SendWindow(branch, ui_mode)
@@ -927,74 +989,74 @@ class cmd_qsend(QBzrCommand):
 
 class cmd_qswitch(QBzrCommand):
     """Set the branch of a checkout and update."""
-    
+
     takes_args = ['location?']
     takes_options = [ui_mode_option]
-    
+
     def _qbzr_run(self, location=None, ui_mode=False):
         from bzrlib.plugins.qbzr.lib.switch import QBzrSwitchWindow
-        
+
         branch = Branch.open_containing(CUR_DIR)[0]
         bzrdir = BzrDir.open_containing(CUR_DIR)[0]
         self.main_window = QBzrSwitchWindow(branch, bzrdir, location, ui_mode)
         self.main_window.show()
-        self._application.exec_() 
+        self._application.exec_()
 
 
 class cmd_qunbind(QBzrCommand):
     """Convert the current checkout into a regular branch."""
     takes_options = [ui_mode_option, execute_option]
-    
+
     def _qbzr_run(self, ui_mode=False, execute=False):
         from bzrlib.plugins.qbzr.lib.unbind import QBzrUnbindDialog
-        
+
         branch = Branch.open_containing(CUR_DIR)[0]
         if branch.get_bound_location() == None:
             raise errors.BzrCommandError("This branch is not bound.")
-        
+
         self.main_window = QBzrUnbindDialog(branch, ui_mode, execute)
         self.main_window.show()
-        self._application.exec_() 
+        self._application.exec_()
 
 
 class cmd_qexport(QBzrCommand):
     """Export current or past revision to a destination directory or archive.
-      
+
       DEST is the destination file or dir where the branch will be exported.
       If BRANCH_OR_SUBDIR is omitted then the branch containing the current working
       directory will be used.
     """
-    
+
     takes_args = ['dest?','branch_or_subdir?']
     takes_options = [ui_mode_option]
-    
+
     def _qbzr_run(self, dest=None, branch_or_subdir=None, ui_mode=False):
         from bzrlib.plugins.qbzr.lib.export import QBzrExportDialog
-        
+
         if branch_or_subdir == None:
             branch = Branch.open_containing(CUR_DIR)[0]
         else:
             branch = Branch.open_containing(branch_or_subdir)[0]
-        
+
         window = QBzrExportDialog(dest, branch, ui_mode)
         window.show()
-        self._application.exec_() 
+        self._application.exec_()
 
 
 class cmd_qbind(QBzrCommand):
     """Convert the current branch into a checkout of the supplied branch.
-    
+
     LOCATION is the branch where you want to bind your current branch.
     """
-    
+
     takes_args = ['location?']
     takes_options = [ui_mode_option]
-    
+
     def _qbzr_run(self, location=None, ui_mode=False):
         from bzrlib.plugins.qbzr.lib.bind import QBzrBindDialog
-        
+
         branch = Branch.open_containing(CUR_DIR)[0]
-        
+
         self.main_window = QBzrBindDialog(branch, location, ui_mode)
         self.main_window.show()
         self._application.exec_()
@@ -1030,7 +1092,11 @@ class cmd_qrun(QBzrCommand):
         directory=None, category=None, execute=False):
         from bzrlib.plugins.qbzr.lib.run import QBzrRunDialog
         if parameters_list:
-            parameters = " ".join(parameters_list)
+            def quote_spaces(s):
+                if " " in s:
+                    s = '"%s"' % s
+                return s
+            parameters = " ".join([quote_spaces(i) for i in parameters_list])
         else:
             parameters = None
         if not command:
@@ -1040,3 +1106,70 @@ class cmd_qrun(QBzrCommand):
             execute=execute)
         window.show()
         self._application.exec_()
+        
+class cmd_qshelve(QBzrCommand):
+    """Shelve selected changes away."""
+    takes_args = ['file*']
+    takes_options = [
+        ui_mode_option,
+        bzr_option('shelve', 'list'),
+        bzr_option('shelve', 'directory'),
+        bzr_option('shelve', 'message'),
+        Option('all', help='Select all changes.'),
+        Option('complete', help='Show complete files.'),
+        Option('ignore-whitespace', short_name='w',
+               help="Ignore whitespace when finding differences.(Only work when --list specified)"),
+        Option('encoding', type=check_encoding,
+               help='Encoding of files content (default: utf-8).'),
+        ]
+
+    def _qbzr_run(self, file_list=None, list=False, directory=None, ui_mode=False, 
+                            complete=False, ignore_whitespace=False, encoding=None,
+                            all=False, message=None):
+        if list:
+            initial_tab = 1
+        else:
+            initial_tab = 0
+        self.main_window = ShelveWindow(file_list=file_list, directory=directory, ui_mode=ui_mode,
+                                initial_tab=initial_tab, complete=complete, 
+                                ignore_whitespace=ignore_whitespace, encoding=encoding,
+                                select_all=all, message=message)
+        self.main_window.show()
+        self._application.exec_()
+
+class cmd_qunshelve(QBzrCommand):
+    """Restore shalved changes."""
+    takes_options = [
+        ui_mode_option,
+        bzr_option('unshelve', 'directory'),
+        Option('complete', help='Show complete files.'),
+        Option('ignore-whitespace', short_name='w',
+               help="Ignore whitespace when finding differences.(Only work when --list specified)"),
+        Option('encoding', type=check_encoding,
+               help='Encoding of files content (default: utf-8).'),
+        ]
+
+    def _qbzr_run(self, directory=None, ui_mode=False, 
+                        complete=False, ignore_whitespace=False, encoding=None):
+        self.main_window = ShelveWindow(directory=directory, ui_mode=ui_mode,
+                                initial_tab=1, complete=complete, 
+                                ignore_whitespace=ignore_whitespace, encoding=encoding)
+        self.main_window.show()
+        self._application.exec_()
+
+
+class cmd_qignore(QBzrCommand):
+    """Ignore files or patterns."""
+    takes_args = []
+    takes_options = [
+        ui_mode_option,
+        bzr_option('ignore', 'directory'),
+        ]
+    aliases = []
+
+    def _qbzr_run(self, directory=None, ui_mode=False):
+        from bzrlib.plugins.qbzr.lib.ignore import IgnoreWindow
+        wt = WorkingTree.open_containing(directory)[0]
+        self.main_window = IgnoreWindow(tree=wt, ui_mode=ui_mode)
+        self.main_window.show()
+        self._application.exec_()
