@@ -27,6 +27,7 @@ from bzrlib.conflicts import TextConflict, ConflictList
 from bzrlib import ignores
 
 from PyQt4 import QtCore, QtGui
+from bzrlib.plugins.qbzr.lib import tests as qtests
 from bzrlib.plugins.qbzr.lib.treewidget import (
     TreeWidget,
     TreeModel,
@@ -37,7 +38,6 @@ from bzrlib.plugins.qbzr.lib.treewidget import (
     group_large_dirs,
     )
 from bzrlib.plugins.qbzr.lib.tests.modeltest import ModelTest
-from bzrlib.plugins.qbzr.lib.tests.excepthookwatcher import TestWatchExceptHook
 
 def load_tests(standard_tests, module, loader):
     result = loader.suiteClass()
@@ -53,23 +53,24 @@ def load_tests(standard_tests, module, loader):
                 TestTreeFilterProxyModel,
                 )))
     tests.multiply_tests(filter_tests, filter_scenarios, result)
-    
+
     # No parametrization for the remaining tests
     result.addTests(remaining_tests)
 
     return result
 
-class TestTreeWidget(TestWatchExceptHook, TestCaseWithTransport):
-    
+
+class TestTreeWidget(qtests.QTestCase):
+
     # Set by load_tests
     make_tree = None
     modify_tree = None
     changes_mode = False
-    
+
     def setUp(self):
         super(TestTreeWidget, self).setUp()
         self.tree, self.branch = self.make_tree(self)
-    
+
     def make_working_tree(self):
         #tree = WorkingTree()
         tree = self.make_branch_and_tree('trunk')
@@ -224,14 +225,15 @@ tree_scenarios = (
          'modify_tree': lambda self, tree: None,}),
 )
 
-class TestTreeFilterProxyModel(TestWatchExceptHook, TestCaseWithTransport):
+class TestTreeFilterProxyModel(qtests.QTestCase):
+
     # Set by load_tests
     filter = None
     expected_visible = None
-    
+
     def test_filters(self):
         tree = self.make_branch_and_tree('tree')
-        
+
         self.build_tree(['tree/dir-with-unversioned/',
                          'tree/ignored-dir-with-child/',])
         self.build_tree_contents([('tree/dir-with-unversioned/child', ''),
@@ -247,26 +249,26 @@ class TestTreeFilterProxyModel(TestWatchExceptHook, TestCaseWithTransport):
         ignores.tree_ignores_add_patterns(tree,
                                           ['ignored-dir-with-child',
                                            'ignored'])
-        
+
         tree.commit('a', rev_id='rev-a',
                     committer="joe@foo.com",
                     timestamp=1166046000.00, timezone=0)
-        
+
         self.build_tree_contents([('tree/changed', 'new')])
-        
+
         self.model = TreeModel()
         load_dirs=[PersistantItemReference(None, 'dir-with-unversioned'),
-                   PersistantItemReference(None, 'ignored-dir-with-child')]        
+                   PersistantItemReference(None, 'ignored-dir-with-child')]
         self.model.set_tree(tree, branch=tree.branch, load_dirs=load_dirs)
         self.filter_model = TreeFilterProxyModel()
         self.filter_model.setSourceModel(self.model)
-        
+
 
         self.filter_model.setFilters(self.filter)
-        self.assertVisiblePaths(self.expected_visible)
+        self.expected_visible.sort()
+        self.assertEqual(self.getVisiblePaths(), self.expected_visible)
 
-    
-    def assertVisiblePaths(self, paths):
+    def getVisiblePaths(self):
         visible_paths = []
         parent_indexes_to_visit = [QtCore.QModelIndex()]
         while parent_indexes_to_visit:
@@ -277,18 +279,41 @@ class TestTreeFilterProxyModel(TestWatchExceptHook, TestCaseWithTransport):
                     str(self.filter_model.data(index, self.model.PATH).toString()))
                 if self.filter_model.hasChildren(index):
                     parent_indexes_to_visit.append(index)
-        
-        # we do not care for the order in this test.
         visible_paths.sort()
-        paths.sort()
-        self.assertEqual(visible_paths, paths)
+        return visible_paths
+
+    def test_unversioned_move_conflict(self):
+        """Test for bug reported as lp:557603 lp:712931 lp:815822 lp:876180"""
+        tree = self.make_branch_and_tree("parent")
+        tree.commit("Base revision")
+        childtree = tree.bzrdir.sprout("child").open_workingtree()
+        self.build_tree(["parent/f", "child/f"])
+        childtree.add(["f"])
+        childtree.commit("Adding f")
+        tree.merge_from_branch(childtree.branch)
+        self.assertLength(1, tree.conflicts())
+        self.assertPathExists("parent/f.moved")
+        os.remove("parent/f.moved")
+        # At this point, the tree has a pending merge adding 'f' and a removed
+        # unversioned duplicate 'f.moved', which is enough to trigger the bug.
+        self.model = TreeModel()
+        load_dirs=[PersistantItemReference(None, "parent")]
+        self.model.set_tree(tree, branch=tree.branch, load_dirs=load_dirs)
+        self.filter_model = TreeFilterProxyModel()
+        self.filter_model.setSourceModel(self.model)
+        self.filter_model.setFilters(self.filter)
+        expected_paths = []
+        if self.filter[TreeFilterProxyModel.CHANGED]:
+            expected_paths.append("f")
+        self.assertEqual(self.getVisiblePaths(), expected_paths)
+
+
+class TestTreeWidgetSelectAll(qtests.QTestCase):
 
-class TestTreeWidgetSelectAll(TestWatchExceptHook, TestCaseWithTransport):
-    
     def setUp(self):
         super(TestTreeWidgetSelectAll, self).setUp()
         tree = self.make_branch_and_tree('tree')
-        
+
         self.build_tree(['tree/dir-with-unversioned/',
                          'tree/ignored-dir-with-child/',
                          'tree/unversioned-with-ignored/',
@@ -308,22 +333,20 @@ class TestTreeWidgetSelectAll(TestWatchExceptHook, TestCaseWithTransport):
         ignores.tree_ignores_add_patterns(tree,
                                           ['ignored-dir-with-child',
                                            'ignored'])
-        
+
         tree.commit('a', rev_id='rev-a',
                     committer="joe@foo.com",
                     timestamp=1166046000.00, timezone=0)
-        
+
         self.build_tree_contents([('tree/changed', 'new')])
         self.tree = tree
-    
+
     def assertSelectedPaths(self, treewidget, paths):
         if 0: treewidget = TreeWidget()
         selected = [item.path for item in treewidget.tree_model.iter_checked()]
         # we do not care for the order in this test.
-        selected.sort()
-        paths.sort()
-        self.assertEqual(selected, paths)
-    
+        self.assertEqual(set(selected), set(paths))
+
     def test_add_selectall(self):
         import bzrlib.plugins.qbzr.lib.add
         self.win = bzrlib.plugins.qbzr.lib.add.AddWindow(self.tree, None)
@@ -333,7 +356,7 @@ class TestTreeWidgetSelectAll(TestWatchExceptHook, TestCaseWithTransport):
                                                      'unversioned',
                                                      'unversioned-with-ignored'])
 
-    
+
     def test_commit_selectall(self):
         import bzrlib.plugins.qbzr.lib.commit
         self.win = bzrlib.plugins.qbzr.lib.commit.CommitWindow(self.tree, None)
@@ -357,7 +380,7 @@ class TestTreeWidgetSelectAll(TestWatchExceptHook, TestCaseWithTransport):
         self.win.initial_load()
         self.win.selectall_checkbox.click()
         self.assertSelectedPaths(self.win.filelist, ['changed'])
-    
+
     def cleanup_win(self):
         # Sometimes the model was getting deleted before the widget, and the
         # widget was trying to query the model. So we delete everything here.
@@ -366,6 +389,7 @@ class TestTreeWidgetSelectAll(TestWatchExceptHook, TestCaseWithTransport):
         self.win.filelist.tree_model.deleteLater()
         QtCore.QCoreApplication.processEvents()
 
+
 class TestModelItemData(TestCase):
 
     def _make_unversioned_model_list(self, iterable):
@@ -408,16 +432,16 @@ filter_scenarios = (
                               'dir-with-unversioned/child',
                               'ignored-dir-with-child',
                               'ignored-dir-with-child/child',
-                              'unchanged', 
-                              'changed', 
-                              'unversioned', 
+                              'unchanged',
+                              'changed',
+                              'unversioned',
                               'ignored',
                               '.bzrignore',
                               ],}),
     ('Unchanged',
         {'filter': (True, False, False, False) ,
          'expected_visible': ['dir-with-unversioned',
-                              'unchanged', 
+                              'unchanged',
                               '.bzrignore',
                               ],}),
     ('Changed',
@@ -427,7 +451,7 @@ filter_scenarios = (
         {'filter': (False, False, True, False),
          'expected_visible': ['dir-with-unversioned',
                               'dir-with-unversioned/child',
-                              'unversioned', 
+                              'unversioned',
                               ],}),
     ('Ignored',
         {'filter': (False, False, False, True),
@@ -441,13 +465,13 @@ filter_scenarios = (
                               'ignored',
                               'dir-with-unversioned',
                               'dir-with-unversioned/child',
-                              'unversioned', 
+                              'unversioned',
                               ],}),
     )
 
 
 class TestGroupLargeDirs(TestCase):
-    
+
     def test_no_group_small(self):
         paths = frozenset(("a/1", "a/2", "a/3", "b"))
         self.assertEqual(group_large_dirs(paths), {"":paths})
@@ -472,13 +496,13 @@ class TestGroupLargeDirs(TestCase):
         paths = frozenset()
         self.assertEqual(group_large_dirs(paths),
                          {'': set([])})
-    
+
     def test_group_deeper_dir(self):
         paths = frozenset(("a/b/1", "a/b/2", "a/b/3", "a/b/4", "c"))
         self.assertEqual(group_large_dirs(paths),
                          {'': set(['a/b', 'c']),
-                          'a/b': set(['a/b/1', 'a/b/2', 'a/b/3', 'a/b/4'])}) 
-    
+                          'a/b': set(['a/b/1', 'a/b/2', 'a/b/3', 'a/b/4'])})
+
     def test_subdir_included(self):
         paths = frozenset([
             u'b',
