@@ -41,7 +41,7 @@ from bzrlib.revisiontree import RevisionTree
 from bzrlib.osutils import file_kind, minimum_path_selection
 from bzrlib.conflicts import TextConflict, resolve
 
-from bzrlib.plugins.qbzr.lib.cat import QBzrCatWindow, QBzrViewWindow
+from bzrlib.plugins.qbzr.lib.cat import QBzrCatWindow, QBzrViewWindow, cat_to_native_app
 from bzrlib.plugins.qbzr.lib.annotate import AnnotateWindow
 from bzrlib.plugins.qbzr.lib.log import LogWindow
 from bzrlib.plugins.qbzr.lib.util import (
@@ -271,7 +271,7 @@ class ChangeDesc(tuple):
     def is_tree_root(desc):
         """Check if entry actually tree root."""
         if desc[3] != (False, False) and desc[4] == (None, None):
-            # TREE_ROOT has not parents (desc[4]).
+            # TREE_ROOT has no parents (desc[4]).
             # But because we could want to see unversioned files
             # we need to check for versioned flag (desc[3])
             return True
@@ -410,7 +410,7 @@ class TreeModel(QtCore.QAbstractItemModel):
         self.inventory_data_by_id = {}
         if is_refresh:
             self.endRemoveRows()
-        
+
         if isinstance(self.tree, WorkingTree):
             tree.lock_read()
             try:
@@ -480,12 +480,12 @@ class TreeModel(QtCore.QAbstractItemModel):
                         else:
                             name = path
                         if change and change.is_renamed():
-                            old_inventory_item = basis_tree.inventory[change.fileid()]
+                            old_inventory_item = self._get_entry(basis_tree, change.fileid())
                             old_names = [old_inventory_item.name]
                             while old_inventory_item.parent_id:
                                 if old_inventory_item.parent_id == dir_fileid:
                                     break
-                                old_inventory_item = basis_tree.inventory[old_inventory_item.parent_id]
+                                old_inventory_item = self._get_entry(basis_tree, old_inventory_item.parent_id)
                                 old_names.append(old_inventory_item.name)
                             old_names.reverse()
                             old_path = "/".join(old_names)
@@ -535,19 +535,20 @@ class TreeModel(QtCore.QAbstractItemModel):
                                 item_data.change)
                 finally:
                     basis_tree.unlock()
-                self.process_inventory(self.working_tree_get_children,
+                self.process_tree(self.working_tree_get_children,
                                        initial_checked_paths, load_dirs)
             finally:
                 tree.unlock()
         else:
-            self.process_inventory(self.revision_tree_get_children,
+            self.process_tree(self.revision_tree_get_children,
                                    initial_checked_paths, load_dirs)
         
         self.emit(QtCore.SIGNAL("layoutChanged()"))
     
     def revision_tree_get_children(self, item_data):
-        for child in item_data.item.children.itervalues():
-            path = self.tree.id2path(child.file_id)
+        for child_id in self.tree.iter_children(item_data.item.file_id):
+            child = self._get_entry(self.tree, child_id)
+            path = self.tree.id2path(child_id)
             yield ModelItemData(path, item=child)
     
     def working_tree_get_children(self, item_data):
@@ -579,19 +580,18 @@ class TreeModel(QtCore.QAbstractItemModel):
                 yield ModelItemData(path, item=child, change=change)
         
         if (not isinstance(item, InternalItem) and
-            item.kind == 'directory' and
-            item.children is not None and not self.changes_mode):
+            item.kind == 'directory' and not self.changes_mode):
             #Because we create copies, we have to get the real item.
-            item = self.tree.inventory[item.file_id]
-            for child in item.children.itervalues():
-                # Create a copy so that we don't have to hold a lock of the wt.
-                child = child.copy()
-                if child.file_id in self.inventory_data_by_id:
-                    child_item_data = self.inventory_data_by_id[child.file_id]
+            item = self._get_entry(self.tree, item.file_id)
+            for child_id in self.tree.iter_children(item.file_id):
+                if child_id in self.inventory_data_by_id:
+                    child_item_data = self.inventory_data_by_id[child_id]
                 else:
-                    path = self.tree.id2path(child.file_id)
+                    path = self.tree.id2path(child_id)
                     child_item_data = ModelItemData(path)
-                
+
+                # Create a copy so that we don't have to hold a lock of the wt.
+                child = self._get_entry(self.tree, child_id).copy()
                 child_item_data.item = child
                 yield child_item_data
         
@@ -617,7 +617,7 @@ class TreeModel(QtCore.QAbstractItemModel):
             return
         dir_item = self.inventory_data[dir_id]
         if dir_item.children_ids is not None:
-            return # This dir has allready been loaded.
+            return # This dir has already been loaded.
         if not dir_item.item.kind=='directory':
             return
         
@@ -641,12 +641,17 @@ class TreeModel(QtCore.QAbstractItemModel):
         finally:
             if not self._many_loaddirs_started:
                 self.tree.unlock()
+
+    def _get_entry(self, tree, file_id):
+        for _, entry in tree.iter_entries_by_dir([file_id]):
+            return entry
+        raise errors.NoSuchId(tree, file_id)
     
-    def process_inventory(self, get_children, initial_checked_paths, load_dirs):
+    def process_tree(self, get_children, initial_checked_paths, load_dirs):
         self.get_children = get_children
         
         root_item = ModelItemData(
-            '', item=self.tree.inventory[self.tree.get_root_id()])
+            '', item=self._get_entry(self.tree, self.tree.get_root_id()))
         
         root_id = self.append_item(root_item, None)
         self.load_dir(root_id)
@@ -1092,11 +1097,11 @@ class TreeModel(QtCore.QAbstractItemModel):
             key = ref.file_id
             dict = self.inventory_data_by_id
             def iter_parents():
-                parent_id = self.tree.inventory[ref.file_id].parent_id
+                parent_id = self._get_entry(self.tree, ref.file_id).parent_id
                 parent_ids = []
                 while parent_id is not None:
                     parent_ids.append(parent_id)
-                    parent_id = self.tree.inventory[parent_id].parent_id
+                    parent_id = self._get_entry(self.tree, parent_id).parent_id
                 return reversed(parent_ids)
         else:
             key = ref.path
@@ -1251,24 +1256,26 @@ class TreeFilterProxyModel(QtGui.QSortFilterProxyModel):
         return False
     
     def filter_id(self, id, item_data):
-        """Determines wether a item should be displayed.
+        """Determines whether a item should be displayed.
         Returns :
             * True: Show the item
-            * False: Donot show the item
+            * False: Do not show the item
             * None: Show the item if there are any children that are visible.
         """
         
         (unchanged, changed, unversioned, ignored) = self.filters
         
-        if item_data.change is None and unchanged: return True
-        
+        is_changed = item_data.change is not None
         is_versioned = item_data.item.file_id is not None
         
-        if is_versioned and item_data.change is not None and changed:
+        if is_versioned and not is_changed and unchanged:
+            return True
+        
+        if is_versioned and is_changed and changed:
             return True
         
         if not is_versioned:
-            if unversioned or ignored:
+            if is_changed and (unversioned or ignored):
                 is_ignored = item_data.change.is_ignored()
                 if not is_ignored and unversioned: return True
                 if is_ignored: return ignored
@@ -1744,7 +1751,6 @@ class TreeWidget(RevisionTreeView):
         single_versioned_file = (single_file and versioned[0])
         
         self.action_open_file.setEnabled(single_item_in_tree)
-        self.action_open_file.setVisible(is_working_tree)
         self.action_show_file.setEnabled(single_file)
         self.action_show_annotate.setEnabled(single_versioned_file)
         self.action_show_log.setEnabled(any(versioned))
@@ -1825,21 +1831,21 @@ class TreeWidget(RevisionTreeView):
     def open_file(self, index=None):
         """Open the file in the os specified editor."""
         
-        if not isinstance(self.tree, WorkingTree):
-            raise AttributeError("Tree must be a working tree to open a file.")
-            
         items = self.get_selection_items([index])
         if not len(items) == 1:
             return
         item = items[0]
         
-        self.tree.lock_read()
-        try:
-            abspath = self.tree.abspath(item.path)
-        finally:
-            self.tree.unlock()
-        url = QtCore.QUrl.fromLocalFile(abspath)
-        QtGui.QDesktopServices.openUrl(url)
+        if isinstance(self.tree, WorkingTree):
+            self.tree.lock_read()
+            try:
+                abspath = self.tree.abspath(item.path)
+            finally:
+                self.tree.unlock()
+            url = QtCore.QUrl.fromLocalFile(abspath)
+            QtGui.QDesktopServices.openUrl(url)
+        else:
+            cat_to_native_app(self.tree, item.path)
 
     @ui_current_widget
     def show_file_log(self):
@@ -1849,7 +1855,7 @@ class TreeWidget(RevisionTreeView):
         fileids = [item.item.file_id for item in items
                    if item.item.file_id is not None]
         
-        window = LogWindow(None, self.branch, fileids)
+        window = LogWindow(branch=self.branch,  specific_file_ids=fileids)
         window.show()
         self.window().windows.append(window)
     
@@ -1862,8 +1868,8 @@ class TreeWidget(RevisionTreeView):
 
         if isinstance(file_id, unicode):
             raise errors.InternalBzrError('file_id should be plain string, not unicode')
-
-        window = AnnotateWindow(self.branch, self.tree, path, file_id)
+        
+        window = AnnotateWindow(self.branch, None, self.tree, path, file_id)
         window.show()
         self.window().windows.append(window)
 
@@ -1886,8 +1892,8 @@ class TreeWidget(RevisionTreeView):
             self.tree.branch, self.tree.branch,
             specific_files=paths)
         
-        show_diff(arg_provider, ext_diff=ext_diff,
-                  parent_window=self.window())
+        show_diff(arg_provider, ext_diff=ext_diff, parent_window=self.window(),
+                  context=self.diff_context)
 
     def unversioned_parents_paths(self, item, include_item=True):
         paths = []
