@@ -29,9 +29,9 @@ from PyQt4 import QtCore, QtGui
 from bzrlib.revision import Revision
 from bzrlib.config import (
     GlobalConfig,
-    IniBasedConfig,
     config_dir,
     ensure_config_dir_exists,
+    config_filename,
     )
 from bzrlib import lazy_regex
 
@@ -50,10 +50,14 @@ lazy_import(globals(), '''
 from bzrlib import (
     osutils,
     urlutils,
+    ui,
 )
-from bzrlib.util.configobj import configobj
 from bzrlib.plugins.qbzr.lib import trace
 from bzrlib.workingtree import WorkingTree
+from bzrlib.transport import get_transport
+from bzrlib.lockdir import LockDir
+
+from bzrlib.plugins.qbzr.lib.compatibility import configobj
 ''')
 
 # standard buttons with translatable labels
@@ -94,22 +98,33 @@ class Config(object):
     def __init__(self, filename):
         self._filename = filename
         self._configobj = None
-
+        dir = osutils.dirname(osutils.safe_unicode(filename))
+        transport = get_transport(dir)
+        self._lock = LockDir(transport, 'lock')
+    
     def _load(self):
         if self._configobj is not None:
             return
         self._configobj = configobj.ConfigObj(self._filename,
                                               encoding='utf-8')
 
-    def setOption(self, name, value, section=None):
+    def set_option(self, name, value, section=None):
         self._load()
         if section is None:
             section = 'DEFAULT'
         if section not in self._configobj:
             self._configobj[section] = {}
-        self._configobj[section][name] = value
+        if value:
+            if not isinstance(value, (str,unicode)):
+                # [bialix 2011/02/11] related to bug #716384: if value is bool
+                # then sometimes configobj lost it in the output file
+                value = str(value)
+            self._configobj[section][name] = value
+        else:
+            if name in self._configobj[section]:
+                del self._configobj[section][name] 
 
-    def getOption(self, name, section=None):
+    def get_option(self, name, section=None):
         self._load()
         if section is None:
             section = 'DEFAULT'
@@ -118,11 +133,19 @@ class Config(object):
         except KeyError:
             return None
 
-    def setSection(self, name, values):
+    def get_option_as_bool(self, name, section=None):
+        # imitate the code from bzrlib.config to read option as boolean
+        # until we will switch to use bzrlib.config instead of our re-implementation
+        value_maybe_str_or_bool = self.get_option(name, section)
+        if value_maybe_str_or_bool not in (None, ''):
+            value = ui.bool_from_string(value_maybe_str_or_bool)
+            return value
+
+    def set_section(self, name, values):
         self._load()
         self._configobj[name] = values
 
-    def getSection(self, name):
+    def get_section(self, name):
         self._load()
         try:
             return self._configobj[name]
@@ -130,11 +153,15 @@ class Config(object):
             return {}
 
     def save(self):
-        self._load()
         ensure_config_dir_exists(os.path.dirname(self._filename))
-        f = open(self._filename, 'wb')
-        self._configobj.write(f)
-        f.close()
+        self._lock.lock_write()
+        try:
+            self._load()
+            f = open(self._filename, 'wb')
+            self._configobj.write(f)
+            f.close()
+        finally:
+            self._lock.unlock()
 
 
 class QBzrConfig(Config):
@@ -142,8 +169,8 @@ class QBzrConfig(Config):
     def __init__(self):
         super(QBzrConfig, self).__init__(config_filename())
 
-    def getBookmarks(self):
-        section = self.getSection('BOOKMARKS')
+    def get_bookmarks(self):
+        section = self.get_section('BOOKMARKS')
         i = 0
         while True:
             try:
@@ -154,19 +181,19 @@ class QBzrConfig(Config):
             i += 1
             yield name, location
 
-    def setBookmarks(self, bookmarks):
+    def set_bookmarks(self, bookmarks):
         section = {}
         for i, (name, location) in enumerate(bookmarks):
             section['bookmark%d' % i] = location
             section['bookmark%d_name' % i] = name
-        self.setSection('BOOKMARKS', section)
+        self.set_section('BOOKMARKS', section)
 
-    def addBookmark(self, name, location):
+    def add_bookmark(self, name, location):
         bookmarks = list(self.getBookmarks())
         bookmarks.append((name, location))
         self.setBookmarks(bookmarks)
         
-    def getColor(self, name, section=None):
+    def get_color(self, name, section=None):
         """
           Get a color entry from the config file.
           Color entries have the syntax:
@@ -194,7 +221,7 @@ class QBzrConfig(Config):
             "Color components for " + name_str +\
             " should be in the range 0..255 only. Given: "+ given +"."
             
-        val = self.getOption(name, section)
+        val = self.get_option(name, section)
         if None == val:
           return None
           
@@ -215,21 +242,37 @@ class QBzrConfig(Config):
         return QtGui.QColor(*color_components)
 
 
-class QBzrGlobalConfig(IniBasedConfig):
+_global_config = None
+def get_global_config():
+    global _global_config
+    
+    if (_global_config is None or
+        _check_global_config_filename_valid(_global_config)):
+        _global_config = GlobalConfig()
+    return _global_config
 
-    def __init__(self):
-        super(QBzrGlobalConfig, self).__init__(config_filename)
+def _check_global_config_filename_valid(config):
+    # before bzr 2.3, there was no file_name attrib, only _get_filename, and
+    # checking that would be meaningless.
+    if hasattr(config, 'file_name'):
+        return not config.file_name == config_filename()
+    else:
+        return False
 
-    def set_user_option(self, option, value):
-        """Save option and its value in the configuration."""
-        conf_dir = os.path.dirname(self._get_filename())
-        ensure_config_dir_exists(conf_dir)
-        if 'DEFAULT' not in self._get_parser():
-            self._get_parser()['DEFAULT'] = {}
-        self._get_parser()['DEFAULT'][option] = value
-        f = open(self._get_filename(), 'wb')
-        self._get_parser().write(f)
-        f.close()
+
+_qbzr_config = None
+def get_qbzr_config():
+    global _qbzr_config
+    if (_qbzr_config is None or
+        not _qbzr_config._filename == config_filename()):
+        _qbzr_config = QBzrConfig()
+    return _qbzr_config
+
+def get_branch_config(branch):
+    if branch: # we should check boolean branch value to support 2 fake branch cases: branch is None, branch is FakeBranch
+        return branch.get_config()
+    else:
+        return get_global_config()
 
 
 class _QBzrWindowBase(object):
@@ -273,7 +316,7 @@ class _QBzrWindowBase(object):
                 QtCore.SIGNAL(signal_name), getattr(self, method_name))
         return buttonbox
 
-    def saveSize(self):
+    def _saveSize(self, config):
         name = self._window_name
         is_maximized = int(self.windowState()) & QtCore.Qt.WindowMaximized != 0
         if is_maximized:
@@ -282,15 +325,18 @@ class _QBzrWindowBase(object):
             size = geom.width(), geom.height()
         else:
             size = self.width(), self.height()
-        config = QBzrGlobalConfig()
-        config.set_user_option(name + "_window_size", "%dx%d" % size)
-        config.set_user_option(name + "_window_maximized", is_maximized)
-        return config
+        config.set_option(name + "_window_size", "%dx%d" % size)
+        config.set_option(name + "_window_maximized", is_maximized)
+    
+    def saveSize(self):
+        config = get_qbzr_config()
+        self._saveSize(config)
+        config.save()
 
     def restoreSize(self, name, defaultSize):
         self._window_name = name
-        config = QBzrGlobalConfig()
-        size = config.get_user_option(name + "_window_size")
+        config = get_qbzr_config()
+        size = config.get_option(name + "_window_size")
         if size:
             size = size.split("x")
             if len(size) == 2:
@@ -308,21 +354,20 @@ class _QBzrWindowBase(object):
             self.resize(size.expandedTo(self.minimumSizeHint()))
         self._restore_size = size
 
-        is_maximized = config.get_user_option(name + "_window_maximized")
-        if is_maximized in ("True", "1"):
+        is_maximized = config.get_option_as_bool(name + "_window_maximized")
+        if is_maximized:
             self.setWindowState(QtCore.Qt.WindowMaximized)
         return config
 
-    def saveSplitterSizes(self):
+    def _saveSplitterSizes(self, config, splitter):
         name = self._window_name
-        config = QBzrGlobalConfig()
-        sizes = ':'.join(map(str, self.splitter.sizes()))
-        config.set_user_option(name + "_splitter_sizes", sizes)
+        sizes = ':'.join(map(str, splitter.sizes()))
+        config.set_option(name + "_splitter_sizes", sizes)
 
     def restoreSplitterSizes(self, default_sizes=None):
         name = self._window_name
-        config = QBzrGlobalConfig()
-        sizes = config.get_user_option(name + "_splitter_sizes")
+        config = get_qbzr_config()
+        sizes = config.get_option(name + "_splitter_sizes")
         n = len(self.splitter.sizes())
         if sizes:
             sizes = map(int, sizes.split(':'))
@@ -361,6 +406,42 @@ class _QBzrWindowBase(object):
     def do_close(self):
         self.close()
 
+    def operation_blocked(self, message):
+        """Use self.operation_blocked in validate methods of q-dialogs
+        to show error message about incorrect or missing parameters.
+
+        We can easily switch between show_error and show_warning
+        inside this method if we want to change the overall qbzr behavior.
+        """
+        self.show_warning(message)
+
+    def show_error(self, message):
+        QtGui.QMessageBox.critical(self,
+            gettext("Error"),
+            message)
+
+    def show_warning(self, message):
+        QtGui.QMessageBox.warning(self,
+            gettext("Warning"),
+            message)
+
+    def ask_confirmation(self, message, type='question'):
+        """Return True if user selected Yes.
+        Optional parameter type selects dialog type. Valid values: question, warning.
+        """
+        klass = QtGui.QMessageBox.question
+        if type == 'warning':
+            klass = QtGui.QMessageBox.warning
+        button = klass(self,
+            gettext("Confirm"),
+            message,
+            QtGui.QMessageBox.Yes | QtGui.QMessageBox.No,
+            QtGui.QMessageBox.No)
+        if button == QtGui.QMessageBox.Yes:
+            return True
+        else:
+            return False
+
 
 class QBzrWindow(QtGui.QMainWindow, _QBzrWindowBase):
 
@@ -520,7 +601,7 @@ def hookup_directory_picker(dialog, chooser, target, chooser_type):
             # Or a QLineEdit
             getter = target.text
             setter = target.setText
-        dir = getter()
+        dir = unicode(getter())
         if not os.path.isdir(dir):
             dir = ""
         dir = QtGui.QFileDialog.getExistingDirectory(dlg, caption, dir)
@@ -530,22 +611,6 @@ def hookup_directory_picker(dialog, chooser, target, chooser_type):
     dialog.connect(chooser, QtCore.SIGNAL("clicked()"), click_handler)
 
 
-_global_config = None
-
-def get_global_config():
-    global _global_config
-    if _global_config is None:
-        _global_config = GlobalConfig()
-    return _global_config
-
-
-def get_branch_config(branch):
-    if branch is not None:
-        return branch.get_config()
-    else:
-        return get_global_config()
-
-
 def open_browser(url):
     try:
         import webbrowser
@@ -607,7 +672,7 @@ def get_set_encoding(encoding, branch):
                 'utf-8 will be used instead') % encoding)
             encoding = 'utf-8'
     else:
-        if branch is not None:
+        if branch: # we should check boolean branch value to support 2 fake branch cases: branch is None, branch is FakeBranch
             branch.get_config().set_user_option("encoding", encoding)
     return encoding
 
@@ -731,13 +796,22 @@ def fill_combo_with(combo, default, *iterables):
             done.add(item)
             combo.addItem(item)
 
+def show_shortcut_hint(action):
+    """Show this action's shortcut, if any, as part of the tooltip.
+    
+    Make sure to set the shortcut and tooltip *before* calling this.
+    """
+    shortcut = action.shortcut()
+    if shortcut and shortcut.toString():
+        toolTip = action.toolTip()
+        action.setToolTip("%s (%s)" % (toolTip, shortcut.toString()))
 
 def iter_saved_pull_locations():
     """ Iterate the 'pull' locations we have previously saved for the user.
     """
-    config = QBzrConfig()
+    config = get_qbzr_config()
     try:
-        sect = config.getSection('Pull Locations')
+        sect = config.get_section('Pull Locations')
     except KeyError:
         return []
     items = sorted(sect.items())
@@ -768,7 +842,7 @@ def save_pull_location(branch, location):
     max_items = 20
     existing = existing[:max_items]
 
-    config = QBzrConfig()
+    config = get_qbzr_config()
     # and save it to the ini
     section = {}
     for i, save_location in enumerate(existing):
@@ -777,7 +851,7 @@ def save_pull_location(branch, location):
         # edits it)
         key = "%04d" % i
         section[key] = save_location
-    config.setSection('Pull Locations', section)
+    config.set_section('Pull Locations', section)
     config.save()
 
 
@@ -847,8 +921,8 @@ loading_queue = None
 def runs_in_loading_queue(f):
     """Methods decorated with this will not run at the same time, but will be
     queued. Methods decorated with this will not be able to return results,
-    but should rather update the ui themselfs. Methods decorated with this
-    should detect, and stop if their execution is no longer requires.
+    but should rather update the ui themselves. Methods decorated with this
+    should detect, and stop if their execution is no longer required.
     
     """
     
@@ -861,13 +935,17 @@ def run_in_loading_queue(cur_f, *cur_args, **cur_kargs):
     global loading_queue
     if loading_queue is None:
         loading_queue = []
-        loading_queue.append((cur_f, cur_args, cur_kargs))
-        
-        while len(loading_queue):
-            f, args, kargs = loading_queue.pop(0)
-            f(*args, **kargs)
-        
-        loading_queue = None
+        try:
+            loading_queue.append((cur_f, cur_args, cur_kargs))
+            
+            while len(loading_queue):
+                try:
+                    f, args, kargs = loading_queue.pop(0)
+                    f(*args, **kargs)
+                except:
+                    trace.report_exception()
+        finally:
+            loading_queue = None
     else:
         loading_queue.append((cur_f, cur_args, cur_kargs))
 
@@ -954,15 +1032,45 @@ def launchpad_project_from_url(url):
 
     @return: project name or None
     """
-    # The format ought to be scheme://host/~user-id/project-name/branch-name/
+    # The format ought to be one of the following:
+    #   scheme://host/~user-id/project-name/branch-name
+    #   scheme://host/+branch/project-name
+    #   scheme://host/+branch/project-name/series-name
+    # there could be distro branches, they are very complex,
+    # so we only support upstream branches based on source package
+    #   scheme://host/+branch/DISTRO/SOURCEPACKAGE
+    #   scheme://host/+branch/DISTRO/SERIES/SOURCEPACKAGE
+    #   scheme://host/+branch/DISTRO/POCKET/SOURCEPACKAGE
+    #   scheme://host/~USER/DISTRO/SERIES/SOURCEPACKAGE/BRANCHNAME
+    DISTROS = ('debian', 'ubuntu')
     from urlparse import urlsplit
     scheme, host, path = urlsplit(url)[:3]
     # Sanity check the host
-    if (host.find('bazaar.launchpad.net') >= 0 or
-        host.find('bazaar.launchpad.dev') >= 0):
+    if (host in ('bazaar.launchpad.net',
+                 'bazaar.launchpad.dev',
+                 'bazaar.qastaging.launchpad.net',
+                 'bazaar.staging.launchpad.net')):
         parts = path.strip('/').split('/')
-        if len(parts) == 3 and parts[0].startswith('~'):
-            return parts[1]
+        if parts[0].startswith('~'):
+            if len(parts) == 3 and parts[1] not in DISTROS:
+                # scheme://host/~user-id/project-name/branch-name/
+                return parts[1]
+            elif len(parts) == 5 and parts[1] in DISTROS:
+                # scheme://host/~USER/DISTRO/SERIES/SOURCEPACKAGE/BRANCHNAME
+                return parts[-2]
+        elif parts[0] in ('%2Bbranch', '+branch'):
+            n = len(parts)
+            if n >= 2:
+                part1 = parts[1]
+                if n in (2,3) and part1 not in DISTROS:
+                    # scheme://host/+branch/project-name
+                    # scheme://host/+branch/project-name/series-name
+                    return part1
+                elif n in (3,4) and part1 in DISTROS:
+                    # scheme://host/+branch/DISTRO/SOURCEPACKAGE
+                    # scheme://host/+branch/DISTRO/SERIES/SOURCEPACKAGE
+                    # scheme://host/+branch/DISTRO/POCKET/SOURCEPACKAGE
+                    return parts[-1]
     return None
 
 
@@ -982,121 +1090,10 @@ if MS_WINDOWS:
 else:
     shlex_split_unicode = _shlex_split_unicode_linux
 
-def get_icon(name):
+def get_icon(name, size=22):
     # TODO: Load multiple sizes
     # TODO: Try load from system theme
-    return QtGui.QIcon(":/22x22/%s.png" % name)
-
-
-class FindToolbar(QtGui.QToolBar):
-    
-    def __init__(self, window, text_edit, show_action):
-        QtGui.QToolBar.__init__(self, gettext("Find"), window)
-        self.text_edit = text_edit
-        if 0: self.text_edit = QtGui.QTextEdit()
-        self.show_action = show_action
-        
-        self.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
-        self.setMovable (False)
-        
-        find_label = QtGui.QLabel(gettext("Find: "), self)
-        self.addWidget(find_label)
-        
-        self.find_text = QtGui.QLineEdit(self)
-        self.addWidget(self.find_text)
-        find_label.setBuddy(self.find_text)
-        
-        prev = self.addAction(get_icon("go-previous"), gettext("Previous"))
-        prev.setShortcut(QtGui.QKeySequence.FindPrevious)
-        
-        next = self.addAction(get_icon("go-next"), gettext("Next"))
-        next.setShortcut(QtGui.QKeySequence.FindNext)
-        
-        self.case_sensitive = QtGui.QCheckBox(gettext("Case sensitive"), self)
-        self.addWidget(self.case_sensitive)
-        self.whole_words = QtGui.QCheckBox(gettext("Whole words"), self)
-        self.addWidget(self.whole_words)
-        
-        close_find = QtGui.QAction(self)
-        close_find.setIcon(self.style().standardIcon(
-                                        QtGui.QStyle.SP_DialogCloseButton))
-        self.addAction(close_find)
-        close_find.setShortcut((QtCore.Qt.Key_Escape))
-        close_find.setShortcutContext(QtCore.Qt.WidgetWithChildrenShortcut)
-        close_find.setStatusTip(gettext("Close find"))
-        self.connect(self.show_action,
-                     QtCore.SIGNAL("toggled (bool)"),
-                     self.show_action_toggle)
-        self.connect(close_find,
-                     QtCore.SIGNAL("triggered(bool)"),
-                     self.close_triggered)
-        self.connect(self.find_text,
-                     QtCore.SIGNAL("textChanged(QString)"),
-                     self.find_text_changed)
-        self.connect(next,
-                     QtCore.SIGNAL("triggered(bool)"),
-                     self.find_next)
-        self.connect(prev,
-                     QtCore.SIGNAL("triggered(bool)"),
-                     self.find_prev)
-        self.connect(self.case_sensitive,
-                     QtCore.SIGNAL("stateChanged(int)"),
-                     self.find_text_changed)
-        self.connect(self.whole_words,
-                     QtCore.SIGNAL("stateChanged(int)"),
-                     self.find_text_changed)
-        self.connect(self.find_text,
-                     QtCore.SIGNAL("returnPressed()"),
-                     self.find_next)        
-        
-    def show_action_toggle(self, state):
-        self.setVisible(state)
-        if state:
-            self.find_text.setFocus()
-        else:
-            self.find_text.setText('')
-    
-    def close_triggered(self, state):
-        self.show_action.setChecked(False)
-    
-    def find_text_changed(self, text):
-        self.find_avoid_moving()
-    
-    def find_get_flags(self):
-        flags = QtGui.QTextDocument.FindFlags()
-        if self.case_sensitive.isChecked():
-            flags = flags | QtGui.QTextDocument.FindCaseSensitively
-        if self.whole_words.isChecked():
-            flags = flags | QtGui.QTextDocument.FindWholeWords
-        return flags
-    
-    def find_avoid_moving(self):
-        self.find(self.text_edit.textCursor().selectionStart(), 0,
-                  self.find_get_flags())
-    
-    def find_next(self, state):
-        self.find(self.text_edit.textCursor().selectionEnd(), 0,
-                  self.find_get_flags())
-    
-    def find_prev(self, state):
-        self.find(self.text_edit.textCursor().selectionStart(),
-                  self.text_edit.document().characterCount(),
-                  self.find_get_flags() | QtGui.QTextDocument.FindBackward)
-    
-    def find(self, from_pos, restart_pos, flags):
-        doc = self.text_edit.document()
-        text = self.find_text.text()
-        cursor = doc.find(text, from_pos, flags)
-        if cursor.isNull():
-            # try again from the restart pos
-            cursor = doc.find(text, restart_pos, flags)
-        if cursor.isNull():
-            cursor = self.text_edit.textCursor()
-            cursor.setPosition(cursor.selectionStart())
-            self.text_edit.setTextCursor(cursor)
-            # Maybe make find_text background red like Firefox?
-        else:
-            self.text_edit.setTextCursor(cursor)
+    return QtGui.QIcon(":/%dx%d/%s.png" % (size, size, name))
 
 
 class InfoWidget(QtGui.QFrame):
@@ -1108,3 +1105,76 @@ class InfoWidget(QtGui.QFrame):
         self.setBackgroundRole(QtGui.QPalette.ToolTipBase) 
         self.setForegroundRole(QtGui.QPalette.ToolTipText)
 
+# Hackish test for monospace. Run bzr qcat lib/util.py to check.
+#888888888888888888888888888888888888888888888888888888888888888888888888888888
+#                                                                             8
+
+monospace_font = None
+def get_monospace_font():
+    global monospace_font
+    if monospace_font is None:
+        monospace_font = _get_monospace_font()
+    return monospace_font
+
+def _get_monospace_font():
+    # TODO: Get font from system settings for Gnome, KDE, Mac.
+    # (no windows option as far as I am aware)
+    # Maybe have our own config setting.
+    
+    # Get the defaul font size
+    size = QtGui.QApplication.font().pointSize()
+    
+    for font_family in ("Monospace", "Courier New"):
+        font = QtGui.QFont(font_family, size)
+        # check that this is really a monospace font
+        if QtGui.QFontInfo(font).fixedPitch():
+            return font
+    
+    # try use style hints to find font.
+    font = QtGui.QFont("", size)
+    font.setFixedPitch(True)
+    return font
+
+def get_set_tab_width_chars(branch=None, tab_width_chars=None):
+    """Function to get the tab width in characters from the configuration.
+
+    @param branch: Use branch.conf as well as bazaar.conf if this is provided.
+    @param tab_width_chars: Number of characters to use as tab width: if branch
+        is provided, the tab width will be stored in branch.conf
+
+    Both arguments are optional, but if tab_width_chars is provided and branch is
+    not, nothing will be done.
+    
+    @return: Tab width, in characters.
+    """
+    if tab_width_chars is None:
+        config = get_branch_config(branch)
+        try:
+            tab_width_chars = int(config.get_user_option('tab_width'))
+            if tab_width_chars < 0:
+                raise TypeError("Invalid tab width")
+        except TypeError:
+            tab_width_chars = 8
+    else:
+        if branch:
+            branch.get_config().set_user_option("tab_width", str(tab_width_chars))
+
+    return tab_width_chars
+
+def get_tab_width_pixels(branch=None, tab_width_chars=None):
+    """Function to get the tab width in pixels based on a monospaced font.
+
+    If tab_width_chars is provided, it is simply converted to a value in pixels.  If
+    it is not provided, the configuration is retrieved from bazaar.conf.  If branch is
+    provided (and tab_width_chars is not), branch.conf is also checked.
+
+    @param tab_width_chars: Number of characters of tab width to convert to pixels.
+    @param branch: Branch to use when retrieving tab width from configuration.
+    
+    @return: Tab width, in pixels.
+    """
+    monospacedFont = get_monospace_font()
+    char_width = QtGui.QFontMetrics(monospacedFont).width(" ")
+    if tab_width_chars is None:
+        tab_width_chars = get_set_tab_width_chars(branch=branch)
+    return char_width*tab_width_chars
